/***
*** pci16AI64.H 
***
***  General description of this file:
***  	Driver structure and macro definitions for General Standards PCI-16AI64
***  	64 channel PCI/PMC A/D board. This file is part of the Linux
***  	driver source distribution for this board.
***  	
***  	This file is not necessary to compile application programs, therefore 
***  	should NOT be included in binary only driver distributions.
***
***  Copyrights (c):
***  	General Standards Corporation (GSC), October 2002
***
***  Author:
***  	Evan Hillman, GSC Inc. (evan@generalStandards.com)
***
***  Support:
***  	Primary support for this driver is provided by GSC. 
***
***  Platform (tested on, may work with others):
***  	Linux, kernel version 2.2.x, Intel hardware.
***
***  Revision history:
***  	Oct  2002: first release
***/

#if   	 LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#define  LINUX_2_4
#elif 	 LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#define  LINUX_2_2
#else
#define  LINUX_2_0
#endif

/* module name(s) */
#define GSC_NAME      "gsc16ai64" /* Must match name in makefile */
#define DRIVER_VERSION  "2.4"

#define GSC_SUBVENDOR PCI_VENDOR_ID_PLX

#define PCI_NUM_BARS 6

#define MAX_BOARDS 10 /* maximum number of boards that driver supports. Arbitrarily chosen,
change if you can actully support more. */

// 12AI64

#define GSE_SUBDEVICE12 0x2406

// 16AI64

#define GSE_SUBDEVICE16 0x2407

struct board_entry {
	long subsystem_vendor;
	long subsystem_device;
	char name[40];
	int index;
};


#if !defined(U64)
typedef union _U64
{
	struct
	{
		u32  LowPart;
		u32  HighPart;
	}u;

	u64 QuadPart;
} U64;
#endif


// PCI BAR Space information 
typedef struct _PCI_BAR_INFO
{
	u32     *pVa;                        // BAR Kernel Virtual Address
	U64      Physical;                   // BAR Physical Address
	u32      Size;                       // BAR size
	int  IsIoMapped;					 // Memory or I/O mapped?
} PCI_BAR_INFO;

/* card descriptor structure */
struct device_board {
	struct pci_dev *pdev;       	/* the kernel PCI device descriptor */
	struct file *fp;				// file pointer for user data.
	u32 *runtime_addr;				/* remmapped runtime register address space */
	u32 *local_addr;				/* remmaped local register address space */
	u32 *intermediateBuffer;   		/* DMA buffer */
	u32 intermediatePhysicalAddr;   /* physical address of above */
	long timeout;      	      		/* timeout flag */
	int timeout_seconds;			/* time to wait before timeout (in seconds) */
	int fillBuffer;					/* Flag to wait for datatime to wait before timeout (in seconds) */

	int dmaState;      	      		/* DMA mode transfer enable flag */
	unsigned long intermediateSamples;       /* unread samples in the DMA buffer */
	u32 intermediateStart;       	/* first unread sample in the DMA buffer */
	u32 userOffset;					// current offset into user buffer.
	unsigned long userRemaining;	// space remaining in the user buffer.
	u32 *userBuffer;				// pointer to the current user buffer.
	int busy;						/* device opened */
	int irqlevel; 					/* interrupt level */
	int minor;						/* device minor number */
	u32 bufferThreshold;
	int detectOverflow;
	int error;						/* SDI error code */
	int signalno; 					/* signal index if user requested notification */
	int signalev;       	      	/* signal the user if this event occurs */
	volatile char init_pending; 	/* flag set if init IRQ expected */
	volatile char autocal_pending; 	/* flag set if autocal IRQ expected */
	volatile char chready_pending; 	/* flag set if channels ready IRQ expected */
	volatile char lo_hi_pending;    /* flag set if waiting for low to high buffer transition */
	volatile char dma_pending; 		/* flag set when DMA is started */
#ifdef LINUX_2_4
	wait_queue_head_t ioctlwq;		/* queue of tasks sleeping in ioctl */
	wait_queue_head_t readwq;		/* queue of tasks sleeping in read */
	wait_queue_head_t dmawq;		/* queue of tasks sleeping on DMA transfer completion */
#else
	struct wait_queue *ioctlwq;     /* queue of tasks sleeping in ioctl */
	struct wait_queue *readwq;      /* queue of tasks sleeping in read */
	struct wait_queue *dmawq;      	/* queue of tasks sleeping on DMA transfer completion */
#endif
	int board_type;					/* an index for which board type for drivers that support multiple types. */
	int board_index;

	struct timer_list watchdog_timer;
	PCI_BAR_INFO           PciBar[PCI_NUM_BARS];    // PCI BARs information

	struct device_board *next;     	    /* next board in linked list */
};

/* non-static function prototypes */
int device_ioctl(struct inode *inode, struct file *fp, unsigned int num, unsigned long arg);
void device_interrupt(int irq, void *dev_id, struct pt_regs *regs);

/* timing and delay stuff */
#define INIT_TIMEOUT_MS       	2000 	    /* initialization timeout */
#define AUTOCAL_TIMEOUT_MS    	5000  	    /* auto calibration timeout */
#define CH_READY_TIMEOUT_MS   	500   	    /* timeout after channel setup operations */
#define MSECS_TO_SLEEP(MS)      ((MS * HZ) / 1000)

/* additional, internally used interrupt event codes */
#define NO_EVENT      	      	(-1)

/* DMA buffer size (16 kBytes on Intel) */
//#define DMA_ORDER     	      	4
#define DMA_ORDER     	      	4
#define DMA_SIZE      	      	(PAGE_SIZE << DMA_ORDER)
#define DMA_SAMPLES   	      	(DMA_SIZE / sizeof(long))

/* DMA command codes */
#define SETUP_DMA_CMD_0         0x09
#define STARTUP_DMA_CMD_0       0x03
#define START_DMA_CMD_1         0xB00
#define STOP_DMA_CMD_0_MASK     0x0C
#define STOP_DMA_CMD_1_MASK     (0x0C<<16)
#define NON_DEMAND_DMA_MODE     0x00020D43
#define DEMAND_DMA_MODE			(0x00020D43|1<<12)
#define DMA_CMD_STAT_INT_CLEAR  0x08
#define PCI_INT_ENABLE        	0x00050900

/* Macros to make register access easier */
/* Mapping Local Configuration Registers */
#define PciLocRange0(card)      ((card)->runtime_addr + PCI_TO_LOC_ADDR_0_RNG)
#define PciLocRemap0(card)      ((card)->runtime_addr + LOC_BASE_ADDR_REMAP_0)
#define ModeArb(card)           ((card)->runtime_addr + MODE_ARBITRATION)
#define EndianDescr(card)       ((card)->runtime_addr + BIG_LITTLE_ENDIAN_DESC)
#define PciLERomRange(card)     ((card)->runtime_addr + PCI_TO_LOC_ROM_RNG)
#define PciLERomRemap(card)     ((card)->runtime_addr + LOC_BASE_ADDR_REMAP_EXP_ROM)
#define PciLBRegDescr0(card)    ((card)->runtime_addr + BUS_REG_DESC_0_FOR_PCI_LOC)
#define LocPciRange(card)       ((card)->runtime_addr + DIR_MASTER_TO_PCI_RNG)
#define LocPciMemBase(card)     ((card)->runtime_addr + LOC_ADDR_FOR_DIR_MASTER_MEM)
#define LocPciIOBase(card)      ((card)->runtime_addr + LOC_ADDR_FOR_DIR_MASTER_IO)
#define LocPciRemap(card)       ((card)->runtime_addr + PCI_ADDR_REMAP_DIR_MASTER)
#define LocPciConfig(card)      ((card)->runtime_addr + PCI_CFG_ADDR_DIR_MASTER_IO)
#define PciLocRange1(card)      ((card)->runtime_addr + PCI_TO_LOC_ADDR_1_RNG)
#define PciLocRemap1(card)      ((card)->runtime_addr + LOC_BASE_ADDR_REMAP_1)
#define PciLBRegDescr1(card)    ((card)->runtime_addr + BUS_REG_DESC_1_FOR_PCI_LOC)

/* Mapping Runtime Registers */
#define Mailbox0(card)          ((card)->runtime_addr + MAILBOX_REGISTER_0)
#define Mailbox1(card)          ((card)->runtime_addr + MAILBOX_REGISTER_1)
#define Mailbox2(card)          ((card)->runtime_addr + MAILBOX_REGISTER_2)
#define Mailbox3(card)          ((card)->runtime_addr + MAILBOX_REGISTER_3)
#define Mailbox4(card)          ((card)->runtime_addr + MAILBOX_REGISTER_4)
#define Mailbox5(card)          ((card)->runtime_addr + MAILBOX_REGISTER_5)
#define Mailbox6(card)          ((card)->runtime_addr + MAILBOX_REGISTER_6)
#define Mailbox7(card)          ((card)->runtime_addr + MAILBOX_REGISTER_7)
#define PciLocDoorBell(card)    ((card)->runtime_addr + PCI_TO_LOC_DOORBELL)
#define LocPciDoorBell(card)    ((card)->runtime_addr + LOC_TO_PCI_DOORBELL)

/* Mapping interrupt control/status reg  */
#define IntCntrlStat(card)      ((card)->runtime_addr + INT_CTRL_STATUS)
#define RunTimeCntrl(card)      ((card)->runtime_addr + PROM_CTRL_CMD_CODES_CTRL)
#define DevVenIDc(card)         ((card)->runtime_addr + DEVICE_ID_VENDOR_ID)
#define RevID(card)             ((card)->runtime_addr + REVISION_ID)
#define MailboxReg0(card)       ((card)->runtime_addr + MAILBOX_REG_0)
#define MailboxReg1(card)       ((card)->runtime_addr + MAILBOX_REG_1)

/* Mapping DMA registers */
#define DMAMode0(card)          ((card)->runtime_addr + DMA_CH_0_MODE)
#define DMAPCIAddr0(card)       ((card)->runtime_addr + DMA_CH_0_PCI_ADDR)
#define DMALocalAddr0(card)     ((card)->runtime_addr + DMA_CH_0_LOCAL_ADDR)
#define DMAByteCnt0(card)       ((card)->runtime_addr + DMA_CH_0_TRANS_BYTE_CNT)
#define DMADescrPtr0(card)      ((card)->runtime_addr + DMA_CH_0_DESC_PTR)
#define DMAMode1(card)          ((card)->runtime_addr + DMA_CH_1_MODE)
#define DMAPCIAddr1(card)       ((card)->runtime_addr + DMA_CH_1_PCI_ADDR)
#define DMALocalAddr1(card)     ((card)->runtime_addr + DMA_CH_1_LOCAL_ADDR)
#define DMAByteCnt1(card)       ((card)->runtime_addr + DMA_CH_1_TRANS_BYTE_CNT)
#define DMADescrPtr1(card)      ((card)->runtime_addr + DMA_CH_1_DESC_PTR)
#define DMACmdStatus(card)      ((card)->runtime_addr + DMA_CMD_STATUS)
#define DMAArbitr(card)         ((card)->runtime_addr + DMA_MODE_ARB_REG)
#define DMAThreshold(card)      ((card)->runtime_addr + DMA_THRESHOLD_REG)

/* Mapping FIFO registers */
#define OutPostQIntStat(card)   ((card)->runtime_addr + OUT_POST_Q_INT_STATUS)
#define OutPostQIntMask(card)   ((card)->runtime_addr + OUT_POST_Q_INT_MASK)
#define MsgUnitCfg(card)        ((card)->runtime_addr + MSG_UNIT_CONFIG)
#define QBaseAddr(card)         ((card)->runtime_addr + Q_BASE_ADDR)
#define InFreeHeadPtr(card)     ((card)->runtime_addr + IN_FREE_HEAD_PTR)
#define InFreeTailPtr(card)     ((card)->runtime_addr + IN_FREE_TAIL_PTR)
#define InPostHeadPtr(card)     ((card)->runtime_addr + IN_POST_HEAD_PTR)
#define InPostTailPtr(card)     ((card)->runtime_addr + IN_POST_TAIL_PTR)
#define OutFreeHeadPtr(card)    ((card)->runtime_addr + OUT_FREE_HEAD_PTR)
#define OutFreeTailPtr(card)    ((card)->runtime_addr + OUT_FREE_TAIL_PTR)
#define OutPostHeadPtr(card)    ((card)->runtime_addr + OUT_POST_HEAD_PTR)
#define OutPostTailPtr(card)    ((card)->runtime_addr + OUT_POST_TAIL_PTR)
#define QStatusCtrl(card)       ((card)->runtime_addr + Q_STATUS_CTRL_REG)

/* Setup all the local registers */
#define BoardCtrlReg(card)      ((card)->local_addr + BOARD_CTRL_REG)
#define IntCtrlReg(card)		((card)->local_addr + INT_CTRL_REG)
#define InputBufferReg(card)    ((card)->local_addr + INPUT_DATA_BUFF_REG)
#define InputCtrlReg(card)      ((card)->local_addr + INPUT_BUFF_CTRL_REG)
#define RateAGenReg(card)		((card)->local_addr + RATE_A_GENERATOR_REG)
#define RateBGenReg(card)		((card)->local_addr + RATE_B_GENERATOR_REG)
#define BufferSizeReg(card)     ((card)->local_addr + BUFFER_SIZE_REG)
#define ScanSynchCtrlReg(card)  ((card)->local_addr + SCAN_SYNCH_CTRL_REG)

void writelocal(struct device_board *,unsigned value, unsigned address);
unsigned readlocal(struct device_board *,unsigned address);
void regdump(struct device_board *device);
int device_read(struct file *fp, char *buf, size_t size, loff_t * lt);
