/***
*** init.c  
***
***  General description of this file:
***     Device driver source code for General Standards 16AI64
***     family of analog input boards. This file is part of the Linux
***     driver source distribution for this board.
***     
***
***  Copyrights (c):
***     General Standards Corporation (GSC), 2004-2005
***
***  Author:
***     Evan Hillman (evan@generalstandards.com)
***
***  Support:
***     Primary support for this driver is provided by GSC. 
***
***/

//////////////////////////////////////////////////////////////////////////
// set the following flag to trace debug messages in this module.
#ifdef DEBUG
#define TRACE_LOCAL TRUE
#endif

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/pci.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/ioctl.h>
#include <linux/version.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "sysdep.h"
#include "gsc16ai64_ioctl.h"
#include "internals.h"

#include "plx_regs.h"

/* local module data */
static struct file_operations device_fops = {
    open:    device_open,
    release: device_close,
    read:    device_read,
    write:   device_write,
    ioctl:   device_ioctl
};

static int device_major = 0;                  /* device major number (dynamically assigned) */
static struct device_board *boards = 0;       /* linked list of all boards found */

__s8    built[32];
int     num_boards;
int     proc_enabled;

// string names for the events for debug logging.

const char *event_name[]=
{
    "EVENT_INIT_COMPLETE",                       
    "EVENT_AUTOCAL_COMPLETE",                    
    "EVENT_CHANNELS_READY",                      
    "EVENT_BUFFER_THRES_LOW_TO_HIGH",            
    "EVENT_BUFFER_THRES_HIGH_TO_LOW",           
    "EVENT_BUFFER_ALMOST_EMPTY",                 
    "EVENT_BUFFER_ALMOST_FULL",                  
    "EVENT_DMA_PENDING",                         
    "EVENT_LAST_IRQ_EVENT",              
};

/************************************************************************/
/* module initialization: detect card(s) and set up descriptor(s)        */
/************************************************************************/
int init_module(void)
{
    int index = 0;
    unsigned short reg;
    struct device_board *device, *devicenext;
    struct pci_dev *pdev=NULL;
    struct proc_dir_entry*  proc;
    int i;
    int found;

    sprintf(built, "%s, %s ", __DATE__, __TIME__);
    printk(KERN_INFO GSC_NAME ": driver ( version: %s) built %s loading on kernel %s\n",  DRIVER_VERSION, built, UTS_RELEASE);
    printk(KERN_INFO GSC_NAME ": Copyright (C) 2001-2004 General Standards Corp. \n");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,74)
    PCI_DEVICE_LOOP(pdev) {
#else
    while((pdev = pci_find_device(PCI_VENDOR_ID_PLX, PCI_ANY_ID, pdev))) {
#endif
#if TRACE_LOCAL
        //msg( ": Checking...vendor: %X device: %X subvendor: %X subsystem: %X\n",pdev->vendor, pdev->device, pdev->subsystem_vendor, pdev->subsystem_device);
#endif
        if ((pdev->vendor == PCI_VENDOR_ID_PLX) && (pdev->device == PCI_DEVICE_ID_PLX_9080)) {

            /* determine if this is one of the boards supported. */

            found=0;
            i=0;
            while(boards_supported[i].subsystem_device != 0){
                if ((boards_supported[i].subsystem_device == pdev->subsystem_device) &&
                    (boards_supported[i].subsystem_vendor == pdev->subsystem_vendor)){
                        found=1;
                        //msg(":     found board %s type %d\n",boards_supported[i].name,i);
                    }
                    if (found) break;
                    i++;
            }

            if (found){
                printk(KERN_INFO GSC_NAME ":     installing board %s type %d\n",boards_supported[i].name,i);
                pci_read_config_word(pdev, 0x2E, &reg);
                //msg(":     config reg=0x%x\n", reg);
                //msg(":     attaching board #%d\n", index + 1);
                device = (struct device_board *)kmalloc(sizeof(struct device_board), GFP_KERNEL);
                memset(device,0,sizeof(device));
                device ->board_type = i;
                if (pci_enable_device(pdev))
                {
                    errmsg(" ERROR - Unable to get interrupt routing\n");
                    kfree(device);
                    return (-ENODEV);
                }
                pci_set_master(pdev);
                {
                    int ResourceCount = 0;

                    for (i = 0; i < PCI_NUM_BARS; ++i)
                    {
                        // Verify the address is valid
                        if (pci_resource_start(
                            pdev,
                            i
                            ) == 0)
                        {
                            continue;
                        }

                        //msg("   Resource %02d\n", ResourceCount);

                        // Increment resource count
                        ResourceCount++;
                        // Get PCI physical address
                        device->PciBar[i].Physical.QuadPart =
                            pci_resource_start(
                            pdev,
                            i
                            );

                        // Determine resource type
                        if (pci_resource_flags(
                            pdev,
                            i
                            ) & IORESOURCE_IO)
                        {
                            //msg("     Type     : I/O Port\n");

                            // Make sure flags are cleared properly
                            device->PciBar[i].Physical.QuadPart &= ~(0x3);
                            device->PciBar[i].IsIoMapped         = TRUE;
                        }
                        else
                        {
                            //msg("     Type     : Memory Space\n");

                            // Make sure flags are cleared properly
                            device->PciBar[i].Physical.QuadPart &= ~(0xf);
                            device->PciBar[i].IsIoMapped         = FALSE;
                        }

                        //msg("     Address  : %08x\n",device->PciBar[i].Physical.u.LowPart);

                        // Get the size
                        device->PciBar[i].Size = 4096; /*
                                                       Plx_pci_resource_len(
                                                       device,
                                                       i
                                                       );*/
                   }
                }

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,0)
                device->runtime_addr = (u32 *)ioremap(pdev->base_address[0], 4096);
                device->ioctlwq = NULL;
                device->readwq = NULL;
                device->dmawq = NULL;
                if (device->PciBar[2].IsIoMapped)
                {
                    //device->local_addr = pdev->resource[2];
                    errmsg(": I/O mapped local registers\n");
                    device->local_addr = base_address[2];
                }
                else
                {
                    device->local_addr = (u32 *)ioremap(pdev->resource[2].start, 4096);
                }
#else
                if (device->PciBar[2].IsIoMapped)
                {
                    device->local_addr = (u32 *)pdev->resource[2].start;
                }
                else
                {
                    device->local_addr = (u32 *)ioremap(pdev->resource[2].start, 4096);
                }
                device->runtime_addr = (u32 *)ioremap(pdev->resource[0].start, pci_resource_len(pdev, 0));
                init_waitqueue_head(&device->ioctlwq);
                init_waitqueue_head(&device->readwq);
                init_waitqueue_head(&device->dmawq);
#endif
                device->pdev = pdev;
                device->fillBuffer=FALSE;
                device->irqlevel = pdev->irq;
                device->busy = 0;
                device->minor = index;
                device->next = boards;
                msg(":     base_address[0]=0x%lX, base_address[2]=0x%lX\n",pdev->resource[0].start, pdev->resource[2].start);
                //msg(":     local_addr=%p, runtime_addr=%p irq=%d\n", device->local_addr, device->runtime_addr, device->irqlevel);
                boards = device;
                device->board_index=index;
#ifdef TRACE_LOCAL
                //context[index]=device;
                //board_type[index]=device->board_type;
#endif
                index++;
                num_boards++;
            }
        }
    }
    if (index == 0) {
        errmsg(": no board found\n");
        return (-ENODEV);
    }
    device_major = register_chrdev(0, GSC_NAME, &device_fops);
    if (device_major < 0) {
        /* OK, could not register -- undo the work we have done above */
        errmsg(": could not register device number\n");
        for (device = boards; device; device = devicenext) {
            devicenext = device->next;
            pci_disable_device(device->pdev);
            iounmap(device->local_addr);
            iounmap(device->runtime_addr);
            kfree(device);
        }
        boards = NULL;
        return (-ENODEV);
    }
    /*
    *   Add /proc file system support.
    */

    if (num_boards)
    {
        remove_proc_entry(GSC_NAME, NULL);  /* Just in case. */
        proc    = create_proc_entry(GSC_NAME, S_IRUGO, NULL);

        if (proc)
        {
            proc_enabled    = 1;
            proc->read_proc = read_proc;
            proc->get_info  = (void*)proc_get_info;
        }
        else
        {
            errmsg( "<1>%s: create_proc_entry() failure.\n",
                GSC_NAME);
            cleanup_module();
        }
    }

#ifdef TRACE_LOCAL
    msg(": major=%d\n", device_major);
    /*{
        int j;
        for (j=0;j<MAX_BOARDS;j++){
            int_other_count[j]=0;
            int_count[j]=0;
            dma_count[j]=0;
            channel_irq[j]=0;
        }
    }*/
#endif
    return 0;
}

/************************************************************************/
/* cleanup when unloading module                                        */
/************************************************************************/
void cleanup_module(void)
{
    struct device_board *device, *devicenext;
    unregister_chrdev(device_major, GSC_NAME);
    for (device = boards; device; device = devicenext) {
        devicenext = device->next;
        iounmap(device->local_addr);
        iounmap(device->runtime_addr);
        pci_disable_device(device->pdev);
        kfree(device);
    }
    if (proc_enabled)   {
        remove_proc_entry(GSC_NAME, NULL);
        proc_enabled    = 0;
    }

    boards = NULL;
    //msg(": unloaded\n");
}

/************************************************************************/
/* open device                                                          */
/************************************************************************/
int device_open(struct inode *inode, struct file *fp)
{
    struct device_board *device;
    u32 regval;
    for (device = boards; device; device = device->next) {
        if (MINOR(inode->i_rdev) == device->minor) {
            if (device->busy) {
                errmsg("(%d): board already opened\n", device->minor);
                return (-EBUSY);
            }
            //msg("(%d): opening board\n", device->minor);
            if (request_irq(device->irqlevel, device_interrupt, SA_SHIRQ, GSC_NAME, device) < 0) {
                errmsg("(%d): cannot get interrupt %d\n", device->minor, device->irqlevel);
                device->busy = 0;
                return (-EBUSY);
            }
            if (!(device->intermediateBuffer = (u32 *)__get_dma_pages(GFP_KERNEL, DMA_ORDER))) {
                errmsg("(%d): can not allocate DMA pages\n", device->minor);
                free_irq(device->irqlevel, device);
                device->busy = 0;
                return (-EBUSY);
            }
#ifdef TRACE_LOCAL
            //memset(device->intermediateBuffer,0xa5,DMA_ORDER*PAGE_SIZE);
#endif
            device->intermediatePhysicalAddr = virt_to_phys((void *)device->intermediateBuffer);
            msg("(%d): DMA virt=%p, phys=0x%x page size %ld size %ld\n", device->minor, device->intermediateBuffer, device->intermediatePhysicalAddr, PAGE_SIZE, DMA_SIZE/*sizeof(device->intermediateBuffer)*/);
            init_timer(&device->watchdog_timer);
            device->timeout_seconds=20;
            device->watchdog_timer.function=timeout_handler;
            device->watchdog_timer.data = (unsigned long) device;
#ifdef DEBUG
            //device->nread=0;
            //device->last=-1;
#endif
            device->busy = 1;
            device->dmaState = DMA_DISABLE;
            device->dmastart = FALSE;
            device->dmasamples = 0;
            memset(device->irq_event_pending,FALSE,sizeof(device->irq_event_pending));
            device->error = DEVICE_SUCCESS;
            device->signalno = (-1);
            device->signalev = NO_EVENT;
            device->intermediateSamples=0;
            device->overflow_threshold = INPUT_BUFFER_SIZE-0x100;
            device->detectOverflow=FALSE;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,0)
            device->ioctlwq = NULL;
            device->readwq = NULL;
            device->dmawq = NULL;
#else
            init_waitqueue_head(&device->ioctlwq);
            init_waitqueue_head(&device->readwq);
            init_waitqueue_head(&device->dmawq);
#endif
            device->event_queue_waiting[EVENT_INIT_COMPLETE]=&device->ioctlwq;
            device->event_queue_waiting[EVENT_AUTOCAL_COMPLETE]=&device->ioctlwq;
            device->event_queue_waiting[EVENT_BUFFER_THRES_LOW_TO_HIGH]=&device->readwq;
            device->event_queue_waiting[EVENT_BUFFER_THRES_HIGH_TO_LOW]=&device->readwq;
            device->event_queue_waiting[EVENT_DMA_PENDING]=&device->dmawq;

            device->hwBufferSize=IDC_INPUT_BUFFER_SIZE;
            device->debug_state=0;

            fp->private_data = device;
            /* reset DMA engines */
            regval = readl(DMACmdStatus(device));
            regval &= STOP_DMA_CMD_0_MASK;
            regval &= STOP_DMA_CMD_1_MASK;
            writel(regval, DMACmdStatus(device));

            // enable bus mastering
            //writel(readl(PciLocRemap0(device))|0x04,PciLocRemap0(device));

            // disable local interrupts 
            DisableIrqLocal(device,0,0);
            msg("(%d): device_open done\n", device->minor);
            MOD_INC_USE_COUNT;
            return (0);
        }
    }
    msg("(%d): can not find board\n", MINOR(inode->i_rdev));
    return (-ENODEV);
}

/************************************************************************/
/* close device                                                         */
/************************************************************************/
int device_close(struct inode *inode, struct file *fp)
{
    struct device_board *device = (struct device_board *)fp->private_data;
    u32 regval;
    //msg("(%d): closing board\n", device->minor);
    DisableIrqPlx(device);
    /* reset DMA engines */
    regval = readl(DMACmdStatus(device));
    regval &= STOP_DMA_CMD_0_MASK;
    regval &= STOP_DMA_CMD_1_MASK;
    writel(regval, DMACmdStatus(device));
    /* reset local IRQ flag */
    DisableIrqLocal(device,0,0);
    /* free resources */
    free_irq(device->irqlevel, device);
    free_pages((u32)device->intermediateBuffer, DMA_ORDER);
    device->busy = 0;
    MOD_DEC_USE_COUNT;
	mark(0x11111151);
    return (0);
}
