/***
 *** pci16AI64_IOCTL.C 
 ***
 ***  General description of this file:
 ***  	Device driver source code for General Standards PCI-16AI64 
 ***  	64 channel PCI A/D board. This file is part of the Linux
 ***  	driver source distribution for this board.
 ***  	
 ***  	This file is not necessary to compile application programs, therefore 
 ***  	should NOT be included in binary only driver distributions.
 ***
 ***  Copyrights (c):
 ***  	General Standards Corporation (GSC), 2002-2003
 ***
 ***  Author:
 ***  	Evan Hillman, GSC Inc. (evan@generalstandards.com)
 ***
 ***  Support:
 ***  	Primary support for this driver is provided by GSC. 
 ***
 ***  Platform (tested on, may work with others):
 ***  	Linux, kernel version 2.4.x, Intel hardware.
 ***
 ***/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/version.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/ioctl.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "sysdep.h"

#include "pci16AI64_ioctl.h"
#include "pci16AI64.h"
#include "plx_regs.h"

/* ioctl file operation: this does the bulk of the work */
int device_ioctl(struct inode *inode, struct file *fp, unsigned int num, unsigned long arg)
{
  struct ai_board *device = (struct ai_board *)fp->private_data;
  struct device_register_params regs;
  unsigned long ulval, regval, ival;
  //printk(KERN_INFO DEVICE_NAME "(%d): device pointer: %p\n", device->minor,device);

  /* main ioctl function dispatch */
  /* 'break' at the end of branches which need to wait for the */
  /* channels ready condition, 'return' from others */
  switch (num) {
  case IOCTL_DEVICE_NO_COMMAND:
    return (0);

  case IOCTL_DEVICE_INIT_BOARD:
	  printk(KERN_INFO DEVICE_NAME "(%d): initializing board\n", device->minor);
#ifdef DEBUG
	  //regdump(device);
#endif

	  device->init_pending = TRUE;
	  /*
	   *	clear out any pending old interrupts.
	   */
	writelocal(device,0, INT_CTRL_REG);
	regval = readl(IntCntrlStat(device));
	regval &= ~PCI_INT_ENABLE;
	writel(regval, (IntCntrlStat(device)));

	  /* enable PLX interrupt. */
	  writel(IRQ_PCI_ENABLE | IRQ_LOCAL_PCI_ENABLE | IRQ_LOCAL_ENABLE, IntCntrlStat(device));

	  /* enable local init complete interrupt. */
	  //writelocal(device,IRQ_0_IDLE|IRQ_0_REQUEST, INT_CTRL_REG);

	  device->timeout=FALSE;
	  device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
	  add_timer(&device->watchdog_timer);

	  writelocal(device,BC_INITIALIZE, BOARD_CTRL_REG);
#ifdef DEBUG
	  //printk("init ->plx int %.8X board control %.8X local int %.8X\n",
		//  readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readlocal(device,INT_CTRL_REG));
#endif
	  wait_event_interruptible(device->ioctlwq,(!device->init_pending));
	  if (device->timeout)
	  {
		  printk(KERN_ERR DEVICE_NAME "(%d): timeout during board initialize\n", device->minor);
		  device->init_pending = FALSE;
		  device->error = DEVICE_IOCTL_TIMEOUT;
		  return (-EIO);
	  }
	  else
		  del_timer_sync(&device->watchdog_timer);

#ifdef DEBUG
	  //regdump(device);
#endif
#ifdef DEBUG
	  //printk("%d): %.8X %.8X %.8X\n",device->minor, readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readl(DMACmdStatus(device)));
#endif
#ifdef DEBUG
	  //printk("plx int %.8X board control %.8X local int %.8X\n",
	//	  readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readlocal(device,INT_CTRL_REG));
#endif

	  //printk(KERN_INFO DEVICE_NAME "(%d): ioctl initalize done\n", device->minor);
	  return (0);

  case IOCTL_DEVICE_START_AUTOCAL:
	  //printk(KERN_INFO DEVICE_NAME "(%d): calibrating board\n", device->minor);
#ifdef DEBUG
	  //printk("plx int %.8X board control %.8X local int %.8X\n",
	//	  readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readlocal(device,INT_CTRL_REG));
#endif
#ifdef DEBUG
	  //regdump(device);
#endif

	  /* enable local interrupt. */
	  regval = readlocal(device,INT_CTRL_REG);
	  regval &= ~IRQ_0_MASK;
	  regval |=	IRQ_0_AUTOCAL_COMPLETE;
	  device->autocal_pending = TRUE;
	  writelocal(device,regval, INT_CTRL_REG);

	  /* start autocal */
	  regval = readlocal(device,BOARD_CTRL_REG);
//	  regval &= ~(IRQ_0_MASK);
	  regval |= BC_AUTOCAL;
	  device->autocal_pending = TRUE;
	  /* enable PLX interrupt. */
	  writel(readl(IntCntrlStat(device)) | IRQ_PCI_ENABLE | IRQ_LOCAL_PCI_ENABLE | IRQ_LOCAL_ENABLE, IntCntrlStat(device));
	  writelocal(device,regval, BOARD_CTRL_REG);

#ifdef DEBUG
	  printk(KERN_INFO DEVICE_NAME "(%d): about to wait calibrating board\n", device->minor);
	 // regdump(device);
#endif
	  /* wait for complete */
	  device->timeout=FALSE;
	  device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
	  add_timer(&device->watchdog_timer);

	  writelocal(device,regval, BOARD_CTRL_REG);

	  wait_event_interruptible(device->ioctlwq,(!device->autocal_pending));

	  if (device->timeout) {
		  printk(KERN_ERR DEVICE_NAME "(%d): timeout when calibrating board\n", device->minor);
		  device->autocal_pending = FALSE;
		  device->error = DEVICE_IOCTL_TIMEOUT;
		  return (-EIO);
	  }
	  else
		  del_timer_sync(&device->watchdog_timer);

#ifdef DEBUG
	  //printk(KERN_INFO DEVICE_NAME "(%d): after wait calibrating board\n", device->minor);
	 // regdump(device);
#endif

	  //printk(KERN_INFO DEVICE_NAME "(%d): ioctl autocal done\n", device->minor);
	  return (0);

  case IOCTL_DEVICE_GET_DEVICE_TYPE:
	  put_user_ret(device->board_type, (unsigned long *)arg, (-EFAULT));
	  return (0);

  case IOCTL_DEVICE_READ_REGISTER:
	//printk(KERN_INFO DEVICE_NAME "(%d): IOCTL_DEVICE_READ_REGISTER\n", device->minor);
    copy_from_user_ret(&regs, (void *)arg, sizeof(regs), (-EFAULT));
    if ((regs.DeviceRegister < BOARD_CTRL_REG) || (regs.DeviceRegister > SCAN_SYNCH_CTRL_REG)) {
      device->error = DEVICE_INVALID_PARAMETER;
      return (-EIO);
    }
//	printk(KERN_INFO DEVICE_NAME "(%d): IOCTL_DEVICE_READ_REGISTER addr %p register %x\n", device->minor,device->local_addr,regs.DeviceRegister);
    regs.ulRegisterValue = readlocal(device, regs.DeviceRegister);
//	printk(KERN_INFO DEVICE_NAME "(%d): IOCTL_DEVICE_READ_REGISTER before copy\n", device->minor);
    copy_to_user_ret((void *)arg, &regs, sizeof(regs), (-EFAULT));
//	printk(KERN_INFO DEVICE_NAME "(%d): IOCTL_DEVICE_READ_REGISTER after copy\n", device->minor);
    return (0);
 
  case IOCTL_DEVICE_WRITE_REGISTER:
    copy_from_user_ret(&regs, (void *)arg, sizeof(regs), (-EFAULT));
    if ((regs.DeviceRegister < BOARD_CTRL_REG) || (regs.DeviceRegister > SCAN_SYNCH_CTRL_REG)) {
      device->error = DEVICE_INVALID_PARAMETER;
      return (-EIO);
    }
    writel(regs.ulRegisterValue, device->local_addr + regs.DeviceRegister);
    return (0);

  case IOCTL_DEVICE_ENABLE_DMA:
	  get_user_ret(ival, (unsigned long *)arg, (-EFAULT));
	  ival = ival ? 1 : 0;
	  device->dmaenable = ival;
	  return (0);

  case IOCTL_DEVICE_GET_DEVICE_ERROR:
	  put_user_ret(device->error, (unsigned long *)arg, (-EFAULT));
	  return (0);

  case IOCTL_DEVICE_INPUT_RANGE_CONFIG:
	get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
    if ((ulval < INPUT_RANGE_2_5) || (ulval > INPUT_RANGE_10)) {
      device->error = DEVICE_INVALID_PARAMETER;
      return (-EIO);
    }
    regval = readlocal(device,BOARD_CTRL_REG);
    regval &= (~(BC_RANGE_MASK));
    regval |= (ulval<<BC_RANGE_SHIFT);
    writelocal(device,regval, BOARD_CTRL_REG);
	return (0);
    break;

  case IOCTL_DEVICE_INPUT_MODE_CONFIG:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < INPUT_DIFFERENTIAL) || (ulval > INPUT_VREF_TEST)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
    regval = readlocal(device,BOARD_CTRL_REG);
    regval &= (~(INPUT_MODE_MASK));
    regval |= (ulval<<INPUT_MODE_SHIFT);
    writelocal(device,regval, BOARD_CTRL_REG);
	return (0);
    break;

  case IOCTL_DEVICE_INPUT_FORMAT_CONFIG:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < FORMAT_TWO_COMPLEMENT) || (ulval > FORMAT_OFFSET_BINARY)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  regval = readlocal(device,BOARD_CTRL_REG);
	  regval &= (~(BC_FORMAT_MASK));
	  regval |= (ulval<<BC_OFFSET_BINARY_SHIFT);
	  writelocal(device,regval, BOARD_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_START_SINGLE_SCAN:
	  regval = readlocal(device,BOARD_CTRL_REG);
	  regval |= BC_INPUT_SYNCH_START;
	  writelocal(device,regval, BOARD_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_SCAN_SIZE:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < SCAN_SINGLE_CHAN) || (ulval > SCAN_64_CHAN)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  regval = readlocal(device,SCAN_SYNCH_CTRL_REG);
	  regval &= (~(SSCR_CHAN_SIZE_MASK));
	  regval |= (ulval);
	  writelocal(device,regval, SCAN_SYNCH_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_SCAN_CLOCK:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < INTERNAL_RATE_A) || (ulval > BCR_INPUT_SYNCH)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  regval = readlocal(device,SCAN_SYNCH_CTRL_REG);

	  regval &= (~(SSCR_SCAN_CLOCK_MASK));
	  regval |= (ulval<<SSCR_CLOCK_SHIFT);
	  writelocal(device,regval, SCAN_SYNCH_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_DISABLE_PCI_INTERRUPTS:
	  writel(0, IntCntrlStat(device));
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_B_CLOCK_SOURCE:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < MASTER_CLOCK) || (ulval > RATE_A_OUTPUT)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  regval = readlocal(device,SCAN_SYNCH_CTRL_REG);

	  regval &= (~(SSCR_RATE_B_CLOCK_SOURCE_MASK));
	  regval |= (ulval<<SSCR_RATE_B_CLOCK_SOURCE_SHIFT);
	  writelocal(device,regval, SCAN_SYNCH_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SINGLE_CHAN_SELECT:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < 0) || (ulval > 63)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  regval = readlocal(device,SCAN_SYNCH_CTRL_REG);

	  regval &= (~(SSCR_CHAN_SIZE_MASK));
	  regval &= (~(SSCR_SINGLE_CHANNEL_MASK));
	  regval |= (ulval<<SSCR_SINGLE_CHANNEL_SHIFT);
	  writelocal(device,regval, SCAN_SYNCH_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_SCAN_RATE_A:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < 0) || (ulval > 0xFFFF)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  regval = readlocal(device,RATE_A_GENERATOR_REG);

	  regval &= (~(RGR_NRATE_MASK));
	  regval |= (ulval);
	  writelocal(device,regval, RATE_A_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_SCAN_RATE_B:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < 0) || (ulval > 0xFFFF)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  regval = readlocal(device,RATE_B_GENERATOR_REG);

	  regval &= (~(RGR_NRATE_MASK));
	  regval |= (ulval);
	  writelocal(device,regval, RATE_B_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_GEN_DISABLE_A:
	  regval = readlocal(device,RATE_A_GENERATOR_REG);
	  regval |= RGR_GEN_DISABLE;
	  writelocal(device,regval, RATE_A_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_GEN_DISABLE_B:
	  regval = readlocal(device,RATE_B_GENERATOR_REG);
	  regval |= RGR_GEN_DISABLE;
	  writelocal(device,regval, RATE_B_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_GEN_ENABLE_A:
	  regval = readlocal(device,RATE_A_GENERATOR_REG);
	  regval &= ~(RGR_GEN_DISABLE);
	  writelocal(device,regval, RATE_A_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_GEN_ENABLE_B:
	  regval = readlocal(device,RATE_B_GENERATOR_REG);
	  regval &= ~(RGR_GEN_DISABLE);
	  writelocal(device,regval, RATE_B_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_BUFFER_THRESHOLD:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < 0) || (ulval > 0xFFFF)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }

	  regval = readlocal(device,INPUT_BUFF_CTRL_REG);
	  regval &= (~(0xFFFF));
	  regval |= (ulval);
	  device->bufferThreshold = regval;
	  writelocal(device,regval, INPUT_BUFF_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_CLEAR_BUFFER:
	  regval = readlocal(device,INPUT_BUFF_CTRL_REG);
	  regval |= IDC_INPUT_CLEAR_BUFFER;
	  writelocal(device,regval, INPUT_BUFF_CTRL_REG);
	  return (0);
	  break;

  default:
	printk(KERN_ERR DEVICE_NAME "(%d): ERROR - unexpected IOCTL\n", device->minor);
    return (-EINVAL);
  }
  return (0);
}
