/***
*** pci16AI64_MAIN.C 
***
***  General description of this file:
***  	Device driver source code for General Standards PCI-16AI64
***  	64 channel PCI A/D board. This file is part of the Linux
***  	driver source distribution for this board.
***  	
***  	This file is not necessary to compile application programs, therefore 
***  	should NOT be included in binary only driver distributions.
***
***  Copyrights (c):
***  	General Standards Corporation (GSC), 2002-2003
***
***  Author:
***  	Evan Hillman, GSC Inc. (evan@generalStandards.com)
***
***  Support:
***  	Primary support for this driver is provided by GSC. 
***
***  Platform (tested on, may work with others):
***  	Linux, kernel version 2.4.x, Intel hardware.
***
***/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/pci.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "sysdep.h"
#include "pci16AI64_ioctl.h"
#include "pci16AI64.h"
#include "plx_regs.h"

MODULE_LICENSE("GPL");

struct board_entry boards_supported[] = {
	/*    subvendor    subsystem name       index ID */
	{GSC_SUBVENDOR,0x2406,"PMC-12AI64",gsc12ai64}, /* PMC-12AI64 */
	{GSC_SUBVENDOR,0x2407,"PMC-16AI64",gsc16ai64}, /* PMC-16AI64 */
	{0,0,"null"}               /* null terminated list.... */
};

/* Driver Version */
static const char device_version[] = "source version - " DRIVER_VERSION " built - " "$Date: " __DATE__ " " __TIME__ " $";

/* module load/unload functions */
int init_module(void);
void cleanup_module(void);

/* local functions */
static int device_write(struct file *, const char *, size_t, loff_t *);
static int device_read(struct file *, char *, size_t, loff_t *);
static int device_open(struct inode *, struct file *);
static int device_close(struct inode *, struct file *);
static int device_read_buffer(struct ai_board *device, char *buf, int nsamp, int noblock);

static int device_read_dma(struct ai_board *device, int nsamples);

/* local module data */
static struct file_operations device_fops = {
open:    device_open,
release: device_close,
read:    device_read,
write:   device_write,
ioctl:   device_ioctl
};

static	__s8	version[8];
static	__s8	built[32];

static int proc_enabled;
static int num_boards;
static int device_major = 0;     	      	  /* device major number (dynamically assigned) */
static struct ai_board *boards = 0;  	  /* linked list of all pci16AI64 boards found */

#ifdef DEBUG
ulong int_other_count[MAX_BOARDS];
ulong int_count[MAX_BOARDS];
ulong dma_count[MAX_BOARDS];
ulong channel_irq[MAX_BOARDS];
ulong channel_expected[MAX_BOARDS];
int board_type[MAX_BOARDS];
void * context[MAX_BOARDS];

void regdump(struct ai_board *device)
{
	int q;

	for (q=0;q<8;q++)
	{
		printk(KERN_INFO DEVICE_NAME ": register %d value: %.8x\n", q, readlocal(device,q));
	}
	printk(KERN_INFO DEVICE_NAME " PLX IRQ %.8X.\n",readl(IntCntrlStat(device)));
}
#endif

/************************************************************************/
/* timeout_handler                                                      */
/************************************************************************/
void timeout_handler(unsigned long ptr)
{
	struct ai_board *device = (struct ai_board *) ptr;

	printk(KERN_INFO DEVICE_NAME "(%d): ERROR - device timeout\n", device->minor);
	device->timeout=TRUE;
#ifdef DEBUG
	printk("timeout -> plx int %.8X board control %.8X local int %.8X DMA status %.8X buffer size: %.8X \n",	
		readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readlocal(device,INT_CTRL_REG),
		readl(DMACmdStatus(device)),readlocal(device,BUFFER_SIZE_REG));

	printk(KERN_INFO DEVICE_NAME "(%d):    init_pending %d\n", device->minor, device->init_pending);
	printk(KERN_INFO DEVICE_NAME "(%d): autocal_pending %d\n", device->minor, device->autocal_pending);
	printk(KERN_INFO DEVICE_NAME "(%d):   lo_hi_pending %d\n", device->minor, device->lo_hi_pending);
	printk(KERN_INFO DEVICE_NAME "(%d): chready_pending %d\n", device->minor, device->chready_pending);
	printk(KERN_INFO DEVICE_NAME "(%d):   dmainprogress %d\n", device->minor, device->dmainprogress);
	printk(KERN_INFO DEVICE_NAME "(%d):   dmainprogress %d\n", device->minor, device->dmainprogress);

#endif

	/* wake up whatever was waiting. */

	if (device->init_pending)
	{
		device->init_pending = FALSE;
		wake_up(&device->ioctlwq);
	}

	if (device->autocal_pending)
	{
		device->autocal_pending = FALSE;
		wake_up(&device->ioctlwq);
	}

	if (device->chready_pending)
	{ 
		device->chready_pending = FALSE;
		wake_up(&device->dmawq);
	}
	if (device->lo_hi_pending)
	{ 
		device->lo_hi_pending = FALSE;
		wake_up(&device->dmawq);
	}
	printk(KERN_INFO DEVICE_NAME "(%d): timeout return\n", device->minor);
}

/******************************************************************************
*
*	Function:	read_proc
*
*	Purpose:
*
*		Implement the read service for the module's /proc file system
*		entry. Read the documentation on this service for details, as
*		we ignore some arguments according to our needs and the
*		documentation.
*
*	Arguments:
*
*		page	The data produced is put here.
*
*		start	Records pointer to where the data in "page" begins.
*
*		offset	The offset into the file where we're to begin.
*
*		count	The limit to the amount of data we can produce.
*
*		eof	Set this flag when we hit EOF.
*
*		data	A private data item that we may use, but don't.
*
*	Returned:
*
*		int	The number of characters written.
*
******************************************************************************/
static int read_proc(
					 char*	page,
					 char**	start,
					 off_t	offset,
					 int	count,
					 int*	eof,
					 void*	data)
{
#define	_PROC_MSG_SIZE	128

#if PAGE_SIZE < _PROC_MSG_SIZE
#error	"/proc file size may be too big."
#endif
	int	i;
#ifdef DEBUG
	int j;
	char	str[256];
#endif

	MOD_INC_USE_COUNT;
	i	= sprintf(page,
		"version: %s\n"
		"built: %s\n"
		"boards: %d\n",
		DRIVER_VERSION,
		built,
		num_boards);
	page[i]	= 0;

#ifdef DEBUG
	strcat(page, "Debug Stats:\n ");
	strcat(page,"type\tother_ints\ttotal_ints\tdma_ints\tchan_irq\tchan_expect\n");

	for(j=0;j<num_boards;j++){
		sprintf(str,"%d\t\t%ld\t\t%ld\t\t%ld\t\t%ld\t\t%ld\t\t%p\n",
			board_type[j],
			int_other_count[j],
			int_count[j],
			dma_count[j],
			channel_irq[j],
			channel_expected[j],
			context[j]
			);
			strcat(page,str);
	}

#endif
	strcat(page, "\n");
	i	= strlen(page) + 1;

	if (i >= PAGE_SIZE) {
		printk(KERN_INFO DEVICE_NAME "/proc/xxx is larger than PAGE_SIZE\n");
		i= PAGE_SIZE-1;
	}

	i--;
	eof[0]	= 1;
	MOD_DEC_USE_COUNT;
	return(i);
}

/******************************************************************************
*
*	Function:	proc_get_info
*
*	Purpose:
*
*		Implement the get_info() service for /proc file system support.
*
*	Arguments:
*
*		page	The data produced is put here.
*
*		start	Records pointer to where the data in "page" begins.
*
*		offset	The offset into the file where we're to begin.
*
*		count	The limit to the amount of data we can produce.
*
*		dummy	A parameter that is unused (for kernel 2.2 only).
*
*	Returned:
*
*		int	The number of characters written.
*
******************************************************************************/
int proc_get_info(
				  char*	page,
				  char**	start,
				  off_t	offset,
				  int	count,
				  int dummy)
{
	int	eof;
	int	i;

	i	= read_proc(page, start, offset, count, &eof, NULL);
	return(i);
}

/************************************************************************/
/* module initalization: detect card(s) and set up descriptor(s)        */
/************************************************************************/
int init_module(void)
{
	struct proc_dir_entry*	proc;
	unsigned short reg;
	struct ai_board *device, *device_next;
	struct pci_dev *pdev;
	int index = 0;
	int i;
	int found;

	int ResourceCount;

	printk(KERN_INFO DEVICE_NAME ": driver (%s) version loading\n", device_version);
	printk(KERN_INFO DEVICE_NAME ": Copyright (C) 2002-2003 General Standards Corp.\n");

	strcpy(version, DRIVER_VERSION);
	sprintf(built, "%s, %s", __DATE__, __TIME__);

#ifdef LINUX_2_4
	pci_for_each_dev(pdev)
#else
	for (pdev = pci_devices; pdev; pdev = pdev->next)
#endif
	{
		//printk(KERN_INFO DEVICE_NAME ": Checking...vendor: %X device: %X subvendor: %X subsystem: %X\n",pdev->vendor, pdev->device, pdev->subsystem_vendor, pdev->subsystem_device);

		if ((pdev->vendor == PCI_VENDOR_ID_PLX) && (pdev->device == PCI_DEVICE_ID_PLX_9080)) {

			/* determine if this is one of the boards supported. */

			found=FALSE;
			i=0;
			while(boards_supported[i].subsystem_device != 0){
				if ((boards_supported[i].subsystem_device == pdev->subsystem_device) &&
					(boards_supported[i].subsystem_vendor == pdev->subsystem_vendor)){
						found=TRUE;
						//printk(KERN_INFO DEVICE_NAME ":     found board %s type %d\n",boards_supported[i].name,i);
					}
					if (found) break;
					i++;
			}

			if (found){
				pci_read_config_word(pdev, 0x2E, &reg);
				//printk(KERN_INFO DEVICE_NAME ": config reg=0x%x\n", reg);
				device = (struct ai_board *)kmalloc(sizeof(struct ai_board), GFP_KERNEL);
				printk(KERN_INFO DEVICE_NAME ": attaching board #%d device pointer %p\n", index + 1,device);
				device->board_type = i;
				/***/

				/*
				 * The bars may be either in I/O space or memory space.  Figure out which
				 * this board is and Deal with either.
				 */
				ResourceCount = 0;

				for (i = 0; i < PCI_NUM_BARS; ++i)
				{
					// Verify the address is valid
					if (pci_resource_start(
						pdev,
						i
						) == 0)
					{
						continue;
					}

					//printk(KERN_INFO DEVICE_NAME "   Resource %02d\n", ResourceCount);

					// Increment resource count
					ResourceCount++;
					// Get PCI physical address
					device->PciBar[i].Physical.QuadPart =
						pci_resource_start(
						pdev,
						i
						);

					// Determine resource type
					if (pci_resource_flags(
						pdev,
						i
						) & IORESOURCE_IO)
					{
						//printk(KERN_INFO DEVICE_NAME "     Type     : I/O Port\n");

						// Make sure flags are cleared properly
						device->PciBar[i].Physical.QuadPart &= ~(0x3);
						device->PciBar[i].IsIoMapped         = TRUE;
					}
					else
					{
						//printk(KERN_INFO DEVICE_NAME "     Type     : Memory Space\n");

						// Make sure flags are cleared properly
						device->PciBar[i].Physical.QuadPart &= ~(0xf);
						device->PciBar[i].IsIoMapped         = FALSE;
					}

					//printk(KERN_INFO DEVICE_NAME "     Address  : %08x\n",device->PciBar[i].Physical.u.LowPart);

					// Get the size - fake it for now, since it is small and fixed.
					device->PciBar[i].Size = 4096; /*
												   Plx_pci_resource_len(
												   device,
												   i
												   );*/
					/*
					if (device->PciBar[i].Size >= (1 << 10))
					{
					printk (KERN_INFO DEVICE_NAME"     Size     : %08x  (%d Kb)\n",	device->PciBar[i].Size, device->PciBar[i].Size >> 10);
					}
					else
					{
					printk(KERN_INFO DEVICE_NAME"     Size     : %08x  (%d bytes)\n",device->PciBar[i].Size, device->PciBar[i].Size);
					}
					*/
				}

#ifdef LINUX_2_4
				if (device->PciBar[2].IsIoMapped)
				{
					device->local_addr = (unsigned long *)pdev->resource[2].start;
				}
				else
				{
					device->local_addr = (unsigned long *)ioremap(pdev->resource[2].start, 4096);
				}
				device->runtime_addr = (unsigned long *)ioremap(pdev->resource[0].start, 4096);

				init_waitqueue_head(&device->ioctlwq);
				init_waitqueue_head(&device->readwq);
				init_waitqueue_head(&device->dmawq);
				//printk(KERN_INFO DEVICE_NAME ": base_address[2]=0x%lx, base_address[0]=0x%lx\n", pdev->resource[2].start, pdev->resource[0].start);
#else
				if (device->PciBar[2].IsIoMapped)
				{
					printk(KERN_ERR DEVICE_NAME ": I/O mapped local registers\n");
					device->local_addr = base_address[2];
				}
				else
				{
					device->local_addr = (unsigned long *)ioremap(pdev->resource[2].start, 4096);
				}
				device->runtime_addr = (unsigned long *)ioremap(pdev->base_address[0], 4096);
				device->ioctlwq = NULL;
				device->readwq = NULL;
				device->dmawq = NULL;
				printk(KERN_INFO DEVICE_NAME ": base_address[2]=0x%lx, base_address[0]=0x%lx\n", pdev->base_address[2], pdev->base_address[0]);
#endif
				/***/
				printk(KERN_INFO DEVICE_NAME ": runtime addr=%p, local addr=%p\n", device->local_addr, device->runtime_addr);
				if (device->local_addr == 0 || device->runtime_addr == 0)
				{
					printk(KERN_ERR DEVICE_NAME ": ERROR - Unable to map memory region\n");
					return (-ENODEV);
				}

				device->pdev = pdev;
				device->irqlevel = pdev->irq;
				device->busy = 0;
				device->minor = index;
				device->next = boards;
#ifdef DEBUG
				//regdump(device);
#endif
				//printk(KERN_INFO DEVICE_NAME ": local_addr=%p, runtime_addr=%p irq=%d\n", device->local_addr, device->runtime_addr, device->irqlevel);
				boards = device;
				index++;
				num_boards++;
			}
		}
	}
	if (index == 0) {
		printk(KERN_INFO DEVICE_NAME ": no board found\n");
		return (-ENODEV);
	}
	device_major = register_chrdev(0, DEVICE_NAME, &device_fops);
	if (device_major < 0) {
		/* OK, could not register -- undo the work we have done above */
		printk(KERN_INFO DEVICE_NAME ": could not register device number\n");
		for (device = boards; device; device = device_next) {
			device_next = device->next;
			iounmap(device->local_addr);
			iounmap(device->runtime_addr);
			kfree(device);
		}
		boards = NULL;
		return (-ENODEV);
	}
	printk(KERN_INFO DEVICE_NAME ": Number of devices %d\n", num_boards);
	/*
	*	Add /proc file system support.
	*/

	if (num_boards)
	{
		remove_proc_entry(DEVICE_NAME, NULL);	/* Just in case. */
		proc	= create_proc_entry(DEVICE_NAME, S_IRUGO, NULL);

		if (proc)
		{
			proc_enabled	= 1;
			proc->read_proc	= read_proc;
			proc->get_info	= (void*) proc_get_info;
		}
		else
		{
			printk(	"<1>%s: create_proc_entry() failure.\n",
				DEVICE_NAME);
			cleanup_module();
		}
	}

	printk(KERN_INFO DEVICE_NAME ": major=%d\n", device_major);
	return 0;
}
/************************************************************************/
/* cleanup when unloading module                                        */
/************************************************************************/
void cleanup_module(void)
{
	struct ai_board *device, *device_next;
	unregister_chrdev(device_major, DEVICE_NAME);
	for (device = boards; device; device = device_next) {
		device_next = device->next;
		iounmap(device->local_addr);
		iounmap(device->runtime_addr);
		kfree(device);
	}
	boards = NULL;
	if (proc_enabled)	{
		remove_proc_entry(DEVICE_NAME, NULL);
		proc_enabled	= 0;
	}

	printk(KERN_INFO DEVICE_NAME ": unloaded\n");
}

/************************************************************************/
/* open device                                                          */
/************************************************************************/
static int device_open(struct inode *inode, struct file *fp)
{
	struct ai_board *device;
	unsigned long regval;
	for (device = boards; device; device = device->next) {
		if (MINOR(inode->i_rdev) == device->minor) {
			if (device->busy) {
				printk(KERN_INFO DEVICE_NAME "(%d): board already opened\n", device->minor);
				return (-EBUSY);
			}
			printk(KERN_INFO DEVICE_NAME "(%d): opening board\n", device->minor);
			if (request_irq(device->irqlevel, device_interrupt, SA_SHIRQ, DEVICE_NAME, device) < 0) {
				printk(KERN_INFO DEVICE_NAME "(%d): can not get interrupt %d\n", device->minor, device->irqlevel);
				device->busy = 0;
				return (-EBUSY);
			}
			if (!(device->dmabuffer = (unsigned long *)__get_dma_pages(GFP_KERNEL, DMA_ORDER))) {
				printk(KERN_INFO DEVICE_NAME "(%d): can not allocate DMA pages\n", device->minor);
				free_irq(device->irqlevel, device);
				device->busy = 0;
				return (-EBUSY);
			}
			device->dmaoffset = virt_to_phys((void *)device->dmabuffer);
			printk(KERN_INFO DEVICE_NAME "(%d): DMA virt=%p, phys=0x%lx\n", device->minor, device->dmabuffer, device->dmaoffset);
			device->busy = 1;
			device->dmaenable = 0;
			device->dmastart = 0;
			device->dmasamples = 0;
			device->dmainprogress = FALSE;
			device->error = DEVICE_SUCCESS;
			device->signalno = (-1);
			device->signalev = NO_EVENT;
			device->init_pending = FALSE;
			device->autocal_pending = FALSE;
			device->chready_pending = FALSE;
			device->lo_hi_pending = FALSE;
			init_timer(&device->watchdog_timer);
			device->timeout_seconds=DEFAULT_TIMEOUT; //default
			device->watchdog_timer.function=timeout_handler;
			device->watchdog_timer.data = (unsigned long) device;
			device->bufferThreshold= 0xFFFE; // device default

#ifdef LINUX_2_4
			init_waitqueue_head(&device->ioctlwq);
			init_waitqueue_head(&device->readwq);
			init_waitqueue_head(&device->dmawq);
#else
			device->ioctlwq = NULL;
			device->readwq = NULL;
			device->dmawq = NULL;
#endif  
			fp->private_data = device;
			/* reset DMA engines */
#ifdef DEBUG
			printk(KERN_INFO DEVICE_NAME "(%d): resetting DMA engines\n", device->minor);
#endif
			regval = 0;
			regval &= STOP_DMA_CMD_0_MASK;
			regval &= STOP_DMA_CMD_1_MASK;
			writel(regval, DMACmdStatus(device));
			/* reset local IRQ flag */
#ifdef DEBUG
			printk(KERN_INFO DEVICE_NAME "(%d): disabling interrupts\n", device->minor);
#endif
			//`			writelocal(device,0, IntCtrlReg(device));
			//writel(0, (IntCntrlStat(device)));
#ifdef DEBUG
			printk(KERN_INFO DEVICE_NAME "(%d): open done\n", device->minor);
#endif
			MOD_INC_USE_COUNT;
#ifdef DEBUG
			//regdump(device);
#endif
			return (0);
		}
	}
	printk(KERN_INFO DEVICE_NAME "(%d): can not find board\n", MINOR(inode->i_rdev));
	return (-ENODEV);
}

/************************************************************************/
/* close device                                                         */
/************************************************************************/
static int device_close(struct inode *inode, struct file *fp)
{
	struct ai_board *device = (struct ai_board *)fp->private_data;
	unsigned long regval;
	printk(KERN_INFO DEVICE_NAME "(%d): closing board\n", device->minor);
	/* disable local interrupts */
	regval = readl(IntCntrlStat(device));
	regval &= (~PCI_INT_ENABLE);
	writel(regval, (IntCntrlStat(device)));
	/* reset DMA engines */
	regval = 0;
	regval &= STOP_DMA_CMD_0_MASK;
	regval &= STOP_DMA_CMD_1_MASK;
	writel(regval, DMACmdStatus(device));
	/* reset local IRQ flag */
	//`	writelocal(device,0, IntCtrlReg(device));
	/* free resources */
	free_irq(device->irqlevel, device);
	free_pages((unsigned long)device->dmabuffer, DMA_ORDER);
	device->busy = 0;
	MOD_DEC_USE_COUNT;
	return (0);
}

/* write operation: easy -- just return an error */
static int device_write(struct file *fp, const char *buf, size_t size, loff_t * lt)
{
	return (-EPERM);
}

/************************************************************************/
/* read operation: either polled or uses PLX DMA on CH0                 */
/************************************************************************/
static int device_read(struct file *fp, char *buf, size_t size, loff_t * lt)
{
	struct ai_board *device = (struct ai_board *)fp->private_data;
	int nsamples = size / sizeof(long);

	/* verify parameters */
	if (nsamples <= 0) {
#ifdef DEBUG
		printk(KERN_INFO DEVICE_NAME "(%d): device_read - zero sample return\n", device->minor);
#endif
		return (0);
	}

	if (!access_ok(VERIFY_WRITE, (unsigned long)buf, nsamples * sizeof(long))) {
#ifdef DEBUG
		printk(KERN_INFO DEVICE_NAME "(%d): device_read - access not OK\n", device->minor);
#endif
		return (-EFAULT);
	}

	/* check whether we have anything left in the buffer */
	if(device->dmasamples > 0) {
		if(nsamples > device->dmasamples) {
			nsamples = device->dmasamples;
		}
		__copy_to_user(buf, &device->dmabuffer[device->dmastart], nsamples * sizeof(long));
		device->dmastart += nsamples;
		device->dmasamples -= nsamples;
		return(nsamples * sizeof(long));
	}
	/* read at most a buffer full at a time */
	if (nsamples > DMA_SAMPLES) {
		nsamples = DMA_SAMPLES;
	}
	/* fill the DMA buffer and satisfy the user request from it */
	return(device_read_buffer(device, buf, nsamples, (fp->f_flags & O_NONBLOCK)));
}

/* Buffered I/O read method */
/* The method first attempts to fill the DMA buffer to a "reasonable" level */
/* then satisfies the user request from it. Any leftover data will be */
/* returned by successive read call(s) */
static int device_read_buffer(struct ai_board *device, char *buf, int nsamples, int noblock)
{
	unsigned long *b;
	int fifosamples, i;

	//printk(KERN_INFO DEVICE_NAME "(%d): device_read_buffer()\n", device->minor);

	/* check how much data we have in the FIFO */

	fifosamples = readlocal(device,BUFFER_SIZE_REG);
	//printk(KERN_INFO DEVICE_NAME "(%d): device_read_buffer first Current fifo samples %d\n", device->minor, fifosamples);
	if (fifosamples < nsamples) // need more
	{
		/* first of all, check for non-blocking I/O */
		if (noblock) {
			printk(KERN_ERR DEVICE_NAME "(%d): device_read_buffer - return error noblock\n", device->minor);
			device->error = DEVICE_SUCCESS;
			return (-EAGAIN);
		}

		// wait for sufficient samples
		//printk(KERN_INFO DEVICE_NAME "(%d): device_read_buffer wait for more data...\n", device->minor);

		writelocal(device,IRQ_1_LO_HI, INT_CTRL_REG);

		device->timeout=FALSE;
		
		device->lo_hi_pending=TRUE;
		device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
		add_timer(&device->watchdog_timer);
		//writel(0x11111111, (Mailbox0(device)));

		writel(readl(IntCntrlStat(device)) | IRQ_PCI_ENABLE | IRQ_LOCAL_PCI_ENABLE, IntCntrlStat(device));
		wait_event_interruptible(device->dmawq,(!device->lo_hi_pending));
		//writel(0x22222222, (Mailbox0(device)));

		if (device->timeout)
		{
			printk(KERN_INFO DEVICE_NAME "(%d): device_read_dma - timeout...\n", device->minor);
			//printk(" (%d): In progress: %d PXL int/ctrl %.8X BC %.8X DMA %.8X\n", device->minor,device->dmainprogress, readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readl(DMACmdStatus(device)));
			device->error = DEVICE_PIO_TIMEOUT;
			return (-EIO);
		}
		else
			del_timer_sync(&device->watchdog_timer);

		if (signal_pending(current)) {
			return (-ERESTARTSYS); 
		}
		fifosamples = readlocal(device,BUFFER_SIZE_REG);
		//printk(KERN_INFO DEVICE_NAME "(%d): device_read_buffer second Current fifo samples %d\n", device->minor, fifosamples);
		if (device->timeout) {
			printk(KERN_ERR DEVICE_NAME "(%d): device_read_buffer - timeout error\n", device->minor);
			device->error = DEVICE_PIO_TIMEOUT;
			return (-EIO);
		}
	}

	// read and return the data.

	//writel(0x66666666, (Mailbox0(device)));
	if (device->dmaenable && fifosamples>0) {
		if (device_read_dma(device, nsamples)!=0)
		{
#ifdef DEBUG
			printk(KERN_ERR DEVICE_NAME "(%d): device_read_buffer - device_read_dma failure\n", device->minor);
#endif
			return (-EAGAIN);
		};
	}
	else {
		for (b = device->dmabuffer, i = fifosamples; --i >= 0; b++) {
			*b = readlocal(device,INPUT_DATA_BUFF_REG);
		}
	}
#ifdef DEBUG
	fifosamples = readlocal(device,BUFFER_SIZE_REG);
	printk(KERN_INFO DEVICE_NAME "(%d): device_read_buffer after xfer fifo samples %d\n", device->minor, fifosamples);
#endif
	if (nsamples >= fifosamples) {
		/* give it all to the user */
		__copy_to_user(buf, device->dmabuffer, fifosamples * sizeof(long));
		//writel(0x77777777, (Mailbox0(device)));
		return (fifosamples * sizeof(long));
	}
	else {
		/* remember some of the data for next time */
		__copy_to_user(buf, device->dmabuffer, nsamples * sizeof(long));
		device->dmasamples = fifosamples - nsamples;
		device->dmastart = nsamples;
		//writel(0x88888888, (Mailbox0(device)));
		return (nsamples * sizeof(long));
	}
	//writel(0x99999999, (Mailbox0(device)));

}

/* fill the buffer using the DMA method */
/* by the time this is called, we know that we have enough samples in the FIFO */
static int device_read_dma(struct ai_board *device, int nsamples)
{
	unsigned long regval;
#ifdef DEBUG
	printk(KERN_INFO DEVICE_NAME "(%d): device_read_dma\n", device->minor);
	//regdump(device);
#endif

	writel(NON_DEMAND_DMA_MODE, DMAMode0(device));
	writel(device->dmaoffset, DMAPCIAddr0(device));
	writel((INPUT_DATA_BUFF_REG << 2), DMALocalAddr0(device));
	writel((nsamples << 2), DMAByteCnt0(device));
	writel(0xA, DMADescrPtr0(device));
	writel(0, DMAArbitr(device));
	writel(0, DMAThreshold(device));
	regval = readl(DMACmdStatus(device));
	regval &= STOP_DMA_CMD_0_MASK;
	writel(regval, DMACmdStatus(device));
	// OK, here we go! 
	device->dmainprogress = TRUE;
	regval = readl(DMACmdStatus(device));
	regval |= SETUP_DMA_CMD_0;
	writel(regval, DMACmdStatus(device));
	regval |= STARTUP_DMA_CMD_0;
	//writel(0xeeeeeeee, (Mailbox0(device)));

	writel(regval, DMACmdStatus(device));

#ifdef DEBUG
	//printk(KERN_INFO DEVICE_NAME "(%d): device_read_dma - about to sleep...\n", device->minor);
#endif

	device->timeout=FALSE;
	device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
	add_timer(&device->watchdog_timer);
	writel(readl(IntCntrlStat(device)) | IRQ_PCI_ENABLE | IRQ_LOCAL_PCI_ENABLE, IntCntrlStat(device));
	wait_event_interruptible(device->dmawq,(!device->dmainprogress));
	//writel(0xffffffff, (Mailbox0(device)));

	if (device->timeout)
	{
		printk(KERN_INFO DEVICE_NAME "(%d): device_read_dma - timeout...\n", device->minor);
		printk(" (%d): In progress: %d PXL int/ctrl %.8X BC %.8X DMA %.8X\n", device->minor,device->dmainprogress, readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readl(DMACmdStatus(device)));

		return (-EAGAIN);
	}
	else
		del_timer_sync(&device->watchdog_timer);

	return 0;
}

/************************************************************************/
/* Spurious Interrupt handler                                           */
/************************************************************************/
#define CLEAR_SPURIOUS_PLX(regval,mask, desc)  { \
	if (regval & mask) \
{ \
	printk(KERN_INFO DEVICE_NAME "(%d): Unexpected interrupt %s %.8lX %.8X %.8X\n",device->minor,desc, regval,mask,~mask); \
	writel((regval & ~mask), IntCntrlStat(device)); \
} \
}
/*printk(KERN_INFO DEVICE_NAME "(%d): Unexpected interrupt %s %.8lX %.8X %.8X\n",device->minor,desc, regval,mask,~mask); \*/

/************************************************************************/
/* Interrupt handler                                                    */
/************************************************************************/
void device_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct ai_board *device = (struct ai_board *)dev_id;
	unsigned long local_irq_reg;
	unsigned long plx_irq_reg;
	unsigned long dma_reg;
	int irq_type;

	//printk(KERN_INFO DEVICE_NAME " Interrupt()\n");
	/* determine if the interrupt is from this board. */

	local_irq_reg = readlocal(device,INT_CTRL_REG);
	plx_irq_reg = readl(IntCntrlStat(device));

	if ((plx_irq_reg & IRQ_PCI_ENABLE) == 0) { /* not ours */
		//printk(KERN_INFO DEVICE_NAME "Interrupt not ours... %.8lx\n",plx_irq_reg);
#ifdef DEBUG
		int_other_count[device->board_index]++;
#endif
		return;
	}
	//printk(KERN_INFO DEVICE_NAME " Interrupt->ours! plx irq: %.8lX local irq: %.8lX plxdma stat: %.8X \n", plx_irq_reg,local_irq_reg,readl(DMACmdStatus(device)));
	//writel(0xaaaa, (Mailbox0(device)));

	plx_irq_reg &= ~PCI_INT_ENABLE;
	writel(plx_irq_reg, (IntCntrlStat(device)));

	/* check for DMA complete interrupt */
	if (device->dmainprogress) {
		dma_reg = readl(DMACmdStatus(device));
		//printk(KERN_INFO DEVICE_NAME "DMA in progress. %.8lX\n",dma_reg);
		if (dma_reg & DMA0_DONE) {
			//writel(0xbbbb, (Mailbox0(device)));
			/* acq interrupt */
			writel(DMA0_CLR_INT, DMACmdStatus(device));
			/* wake up process blocked in 'read()' */
			device->dmainprogress = FALSE;
			wake_up(&device->dmawq);
			//printk(KERN_INFO DEVICE_NAME "DMA end return plx irq %.8X local irq %.8X. plxdma stat%.8X.\n",readl(IntCntrlStat(device)),readlocal(device,INT_CTRL_REG),readl(DMACmdStatus(device)));
			//writel(0xdddd, (Mailbox0(device)));
			return;
		}
	}
	/*
	*	handle group 0 interrupts.
	*/
	//printk(KERN_INFO DEVICE_NAME "before Group 0 interrupt, local irq reg %.8lx\n",local_irq_reg);

	if (local_irq_reg&IRQ_0_REQUEST)
	{
		irq_type=local_irq_reg&IRQ_0_MASK;
		//writelocal(device,local_irq_reg & (~(IRQ_0_REQUEST|IRQ_0_MASK)), INT_CTRL_REG);
		writelocal(device,0, INT_CTRL_REG);
		//printk(KERN_INFO DEVICE_NAME "Group 0 interrupt. type: %x\n",irq_type);

		switch(irq_type) {
	case IRQ_0_IDLE:
		//printk(KERN_INFO DEVICE_NAME ": group 0 IRQ_0_IDLE Interrupt %x\n",irq_type);
		if (device->init_pending) {
			device->init_pending = FALSE;
			wake_up(&device->ioctlwq);
		}
		break;
	case IRQ_0_AUTOCAL_COMPLETE:
		//printk(KERN_INFO DEVICE_NAME ": group 0 IRQ_0_AUTOCAL_COMPLETE Interrupt %x\n",irq_type);
		if (device->autocal_pending) {
			device->autocal_pending = FALSE;
			wake_up(&device->ioctlwq);
		}
		break;
	case IRQ_0_INPUT_SCAN_INIT:
		//printk(KERN_INFO DEVICE_NAME ": group 0 IRQ_0_INPUT_SCAN_INIT Interrupt %x\n",irq_type);
		if (device->init_pending) {
			device->init_pending = FALSE;
			wake_up(&device->ioctlwq);
		}
		break;
	case IRQ_0_INPUT_SCAN_COMPLETE:
#ifdef DEBUG
		printk(KERN_INFO DEVICE_NAME "(%d): channels ready IRQ\n", device->minor);
#endif
		if (device->chready_pending) {
#ifdef DEBUG
			printk(KERN_INFO DEVICE_NAME "(%d): channels ready IRQ was waited for\n", device->minor);
#endif
			device->chready_pending = FALSE;
			wake_up(&device->dmawq);
		}
		break;
	default:
		printk(KERN_INFO DEVICE_NAME ": ERROR - unknown group 0 Interrupt\n");
		break;
		}
	}
	/*
	*	handle group 1 interrupts.
	*/
	if (local_irq_reg&IRQ_1_REQUEST)
	{
		//printk(KERN_INFO DEVICE_NAME "Group 1 interrupt.\n");
	//	writelocal(device,local_irq_reg & (~(IRQ_1_REQUEST|IRQ_1_MASK)), INT_CTRL_REG);
		writelocal(device,0, INT_CTRL_REG);
		irq_type=local_irq_reg&IRQ_1_MASK;

		switch(irq_type) {
	case IRQ_1_LO_HI:
		//printk(KERN_INFO DEVICE_NAME ": IRQ_1_LO_HI 1 Interrupt %x\n",irq_type);
		if (device->lo_hi_pending) {
#ifdef DEBUG
			printk(KERN_INFO DEVICE_NAME "(%d): channels ready IRQ was waited for\n", device->minor);
#endif
			device->lo_hi_pending = FALSE;
			wake_up(&device->dmawq);
		} 
		break;
	case IRQ_1_HI_LO:
		//printk(KERN_INFO DEVICE_NAME ": IRQ_1_HI_LO 1 Interrupt %x\n",irq_type);
		break;
		}
	}
	/*
	*	check every possible interrupt source, to ensure that the
	* current interrupt is turned off, expected or not.
	*/
	/*
	printk(KERN_INFO DEVICE_NAME "spurious check.\n");

	// PLX interrupt-control registers.
	CLEAR_SPURIOUS_PLX(plx_irq_reg,IRQ_LOCAL_LSERR_ABORT, "IRQ_LOCAL_LSERR_ABORT*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_LSERR_ABORT, "IRQ_LOCAL_LSERR_ABORT*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_LSERR_OVERFLOW, "IRQ_LOCAL_LSERR_OVERFLOW*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_GENERATE_SERR, "IRQ_GENERATE_SERR*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_MAILBOX_ENABLE, "IRQ_MAILBOX_ENABLE*");
	//	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_PCI_ENABLE, "IRQ_PCI_ENABLE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_PCI_DOORBELL_ENABLE, "IRQ_PCI_DOORBELL_ENABLE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_ABORT_ENABLE, "IRQ_ABORT_ENABLE*");
	//	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_PCI_ENABLE, "IRQ_LOCAL_PCI_ENABLE*");
	//	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_ENABLE, "IRQ_LOCAL_ENABLE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_DOORBELL_ENABLE, "IRQ_LOCAL_DOORBELL_ENABLE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_DMA_1_ENABLE,"IRQ_DMA_1_ENABLE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_LOCAL_DOORBELL_ACTIVE,"IRQ_LOCAL_DOORBELL_ACTIVE*");
	CLEAR_SPURIOUS_PLX(plx_irq_reg, IRQ_DMA_1_ACTIVE,"IRQ_DMA_1_ACTIVE*");
	*/

	/* if something is pending, leave with the interrupt enabled. */

	if (device->dmainprogress || device->autocal_pending || device->chready_pending || 
		device->lo_hi_pending || device->init_pending)
	{
		writel(readl(IntCntrlStat(device)) | IRQ_PCI_ENABLE | IRQ_LOCAL_PCI_ENABLE, IntCntrlStat(device));
	}

	//printk(KERN_INFO DEVICE_NAME "IRQ end return plx irq %.8X local irq %.8X plxdma stat%.8X.\n",readl(IntCntrlStat(device)),readlocal(device, INT_CTRL_REG),readl(DMACmdStatus(device)));
	//writel(0xcccc, (Mailbox0(device)));

	return;
}

/************************************************************************/
/* writelocal                                                           */
/*                                                                      */
/* Write to the board local registers, whether the local registers      */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
void writelocal(struct ai_board *device, unsigned value, unsigned address)
{
	if (device->PciBar[2].IsIoMapped) /* i/o space */
	{
		//printk(KERN_INFO DEVICE_NAME "writelocal I/O space addr: %.8X value: %.8X.\n",address,value);
		outl(value, (unsigned) device->local_addr+address*4);
	}
	else
	{
		writel(value, device->local_addr+address);
	}
};

/************************************************************************/
/* readlocal                                                            */
/*                                                                      */
/* Read from the board local registers, whether the local registers     */
/* are in I/O space or memory space.                                    */
/*                                                                      */
/*                                                                      */
/************************************************************************/
unsigned readlocal(struct ai_board *device, unsigned address)
{
	if (device->PciBar[2].IsIoMapped) /* i/o space */
	{
		//		printk(KERN_INFO DEVICE_NAME "readlocal I/O space addr: %.8X \n",address);
		return inl((unsigned)device->local_addr+address*4);
	}
	else
	{
		return readl(device->local_addr+address);
	}
	return 0;
};
